""" The models package handles storing of model updates durign the federated training process.
The functionality is used by the combiner service during aggregation of model updates from clients.
By implementing the interface in the base class modelstorage.py, a developer may customize the
behaviour of the framework. """
