""" Module handling storage of objects in S3-compatible object storage. This functionality is used by the controller
to store global models in the model trail in persistent storage. Currently implemented for MinIO, but a '
developer can extend the framwork by implemeting the interface in base.py. """
